@@ -10,14 +10,14 @@ import sys
 import textwrap
 import types
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, Any, AnyStr, ForwardRef, NewType, TypeVar, get_type_hints
+from typing import TYPE_CHECKING, Any, AnyStr, ForwardRef, NewType, TypeVar, Union, get_type_hints
 
 from docutils import nodes
-from docutils.frontend import OptionParser
+from docutils.frontend import get_default_settings
 from sphinx.ext.autodoc.mock import mock
 from sphinx.parsers import RSTParser
 from sphinx.util import logging, rst
-from sphinx.util.inspect import TypeAliasForwardRef, TypeAliasNamespace, stringify_signature
+from sphinx.util.inspect import TypeAliasForwardRef, stringify_signature
 from sphinx.util.inspect import signature as sphinx_signature
 
 from ._parser import parse
@@ -36,7 +36,23 @@ if TYPE_CHECKING:
     from sphinx.ext.autodoc import Options
 
 _LOGGER = logging.getLogger(__name__)
-_PYDATA_ANNOTATIONS = {"Any", "AnyStr", "Callable", "ClassVar", "Literal", "NoReturn", "Optional", "Tuple", "Union"}
+_PYDATA_ANNOTS_TYPING = {"Any", "AnyStr", "Callable", "ClassVar", "Literal", "NoReturn", "Optional", "Tuple", "Union"}
+_PYDATA_ANNOTS_TYPES = {
+    *("AsyncGeneratorType", "BuiltinFunctionType", "BuiltinMethodType"),
+    *("CellType", "ClassMethodDescriptorType", "CoroutineType"),
+    "EllipsisType",
+    *("FrameType", "FunctionType"),
+    *("GeneratorType", "GetSetDescriptorType"),
+    "LambdaType",
+    *("MemberDescriptorType", "MethodDescriptorType", "MethodType", "MethodWrapperType"),
+    # NoneType is special, but included here for completeness' sake
+    *("NoneType", "NotImplementedType"),
+    "WrapperDescriptorType",
+}
+_PYDATA_ANNOTATIONS = {
+    *(("typing", n) for n in _PYDATA_ANNOTS_TYPING),
+    *(("types", n) for n in _PYDATA_ANNOTS_TYPES),
+}
 
 # types has a bunch of things like ModuleType where ModuleType.__module__ is
 # "builtins" and ModuleType.__name__ is "module", so we have to check for this.
@@ -45,6 +61,11 @@ _TYPES_DICT = {getattr(types, name): name for name in types.__all__}
 _TYPES_DICT[types.FunctionType] = "FunctionType"
 
 
+class MyTypeAliasForwardRef(TypeAliasForwardRef):
+    def __or__(self, value: Any) -> Any:
+        return Union[self, value]  # noqa: UP007
+
+
 def _get_types_type(obj: Any) -> str | None:
     try:
         return _TYPES_DICT.get(obj)
@@ -64,7 +85,7 @@ def get_annotation_module(annotation: Any) -> str:
         return "builtins"
     if _get_types_type(annotation) is not None:
         return "types"
-    is_new_type = sys.version_info >= (3, 10) and isinstance(annotation, NewType)
+    is_new_type = isinstance(annotation, NewType)
     if (
         is_new_type
         or isinstance(annotation, TypeVar)
@@ -156,11 +177,11 @@ def get_annotation_args(annotation: Any, module: str, class_name: str) -> tuple[
     return () if len(result) == 1 and result[0] == () else result  # type: ignore[misc]
 
 
-def format_internal_tuple(t: tuple[Any, ...], config: Config) -> str:
-    # An annotation can be a tuple, e.g., for nptyping:
+def format_internal_tuple(t: tuple[Any, ...], config: Config, *, short_literals: bool = False) -> str:
+    # An annotation can be a tuple, e.g., for numpy.typing:
     # In this case, format_annotation receives:
     # This solution should hopefully be general for *any* type that allows tuples in annotations
-    fmt = [format_annotation(a, config) for a in t]
+    fmt = [format_annotation(a, config, short_literals=short_literals) for a in t]
     if len(fmt) == 0:
         return "()"
     if len(fmt) == 1:
@@ -180,12 +201,13 @@ def fixup_module_name(config: Config, module: str) -> str:
     return module
 
 
-def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PLR0911, PLR0912, PLR0915, PLR0914
+def format_annotation(annotation: Any, config: Config, *, short_literals: bool = False) -> str:  # noqa: C901, PLR0911, PLR0912, PLR0915, PLR0914
     """
     Format the annotation.
 
     :param annotation:
     :param config:
+    :param short_literals: Render :py:class:`Literals` in PEP 604 style (``|``).
     :return:
     """
     typehints_formatter: Callable[..., str] | None = getattr(config, "typehints_formatter", None)
@@ -206,7 +228,7 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
         return format_internal_tuple(annotation, config)
 
     if isinstance(annotation, TypeAliasForwardRef):
-        return str(annotation)
+        return annotation.name
 
     try:
         module = get_annotation_module(annotation)
@@ -219,7 +241,7 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
     full_name = f"{module}.{class_name}" if module != "builtins" else class_name
     fully_qualified: bool = getattr(config, "typehints_fully_qualified", False)
     prefix = "" if fully_qualified or full_name == class_name else "~"
-    role = "data" if module == "typing" and class_name in _PYDATA_ANNOTATIONS else "class"
+    role = "data" if (module, class_name) in _PYDATA_ANNOTATIONS else "class"
     args_format = "\\[{}]"
     formatted_args: str | None = ""
 
@@ -233,12 +255,12 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
     # Some types require special handling
     if full_name == "typing.NewType":
         args_format = f"\\(``{annotation.__name__}``, {{}})"
-        role = "class" if sys.version_info >= (3, 10) else "func"
+        role = "class"
     elif full_name in {"typing.TypeVar", "typing.ParamSpec"}:
         params = {k: getattr(annotation, f"__{k}__") for k in ("bound", "covariant", "contravariant")}
         params = {k: v for k, v in params.items() if v}
         if "bound" in params:
-            params["bound"] = f" {format_annotation(params['bound'], config)}"
+            params["bound"] = f" {format_annotation(params['bound'], config, short_literals=short_literals)}"
         args_format = f"\\(``{annotation.__name__}``{', {}' if args else ''}"
         if params:
             args_format += "".join(f", {k}={v}" for k, v in params.items())
@@ -259,20 +281,22 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
                 args_format = f"\\[:py:data:`{prefix}typing.Union`\\[{{}}]]"
                 args = tuple(x for x in args if x is not type(None))
     elif full_name in {"typing.Callable", "collections.abc.Callable"} and args and args[0] is not ...:
-        fmt = [format_annotation(arg, config) for arg in args]
+        fmt = [format_annotation(arg, config, short_literals=short_literals) for arg in args]
         formatted_args = f"\\[\\[{', '.join(fmt[:-1])}], {fmt[-1]}]"
     elif full_name == "typing.Literal":
+        if short_literals:
+            return f"\\{' | '.join(f'``{arg!r}``' for arg in args)}"
         formatted_args = f"\\[{', '.join(f'``{arg!r}``' for arg in args)}]"
     elif is_bars_union:
-        return " | ".join([format_annotation(arg, config) for arg in args])
+        return " | ".join([format_annotation(arg, config, short_literals=short_literals) for arg in args])
 
     if args and not formatted_args:
         try:
             iter(args)
         except TypeError:
-            fmt = [format_annotation(args, config)]
+            fmt = [format_annotation(args, config, short_literals=short_literals)]
         else:
-            fmt = [format_annotation(arg, config) for arg in args]
+            fmt = [format_annotation(arg, config, short_literals=short_literals) for arg in args]
         formatted_args = args_format.format(", ".join(fmt))
 
     escape = "\\ " if formatted_args else ""
@@ -383,7 +407,12 @@ def process_signature(  # noqa: C901, PLR0913, PLR0917
         elif what == "method":
             # bail if it is a local method as we cannot determine if first argument needs to be deleted or not
             if "<locals>" in obj.__qualname__ and not _is_dataclass(name, what, obj.__qualname__):
-                _LOGGER.warning('Cannot handle as a local function: "%s" (use @functools.wraps)', name)
+                _LOGGER.warning(
+                    'Cannot handle as a local function: "%s" (use @functools.wraps)',
+                    name,
+                    type="sphinx_autodoc_typehints",
+                    subtype="local_function",
+                )
                 return None
             outer = inspect.getmodule(obj)
             for class_name in obj.__qualname__.split(".")[:-1]:
@@ -417,18 +446,17 @@ def _is_dataclass(name: str, what: str, qualname: str) -> bool:
 
 
 def _future_annotations_imported(obj: Any) -> bool:
-    _annotations = getattr(inspect.getmodule(obj), "annotations", None)
-    if _annotations is None:
+    annotations_ = getattr(inspect.getmodule(obj), "annotations", None)
+    if annotations_ is None:
         return False
 
     # Make sure that annotations is imported from __future__ - defined in cpython/Lib/__future__.py
     # annotations become strings at runtime
-    future_annotations = 0x100000 if sys.version_info[0:2] == (3, 7) else 0x1000000
-    return bool(_annotations.compiler_flag == future_annotations)
+    return bool(annotations_.compiler_flag == 0x1000000)  # pragma: no cover # noqa: PLR2004
 
 
 def get_all_type_hints(
-    autodoc_mock_imports: list[str], obj: Any, name: str, localns: TypeAliasNamespace
+    autodoc_mock_imports: list[str], obj: Any, name: str, localns: dict[Any, MyTypeAliasForwardRef]
 ) -> dict[str, Any]:
     result = _get_type_hint(autodoc_mock_imports, name, obj, localns)
     if not result:
@@ -477,7 +505,9 @@ def _execute_guarded_code(autodoc_mock_imports: list[str], obj: Any, module_code
                     with mock(autodoc_mock_imports):
                         exec(guarded_code, getattr(obj, "__globals__", obj.__dict__))  # noqa: S102
         except Exception as exc:  # noqa: BLE001
-            _LOGGER.warning("Failed guarded type import with %r", exc)
+            _LOGGER.warning(
+                "Failed guarded type import with %r", exc, type="sphinx_autodoc_typehints", subtype="guarded_import"
+            )
 
 
 def _resolve_type_guarded_imports(autodoc_mock_imports: list[str], obj: Any) -> None:
@@ -499,7 +529,9 @@ def _resolve_type_guarded_imports(autodoc_mock_imports: list[str], obj: Any) ->
             _execute_guarded_code(autodoc_mock_imports, obj, module_code)
 
 
-def _get_type_hint(autodoc_mock_imports: list[str], name: str, obj: Any, localns: TypeAliasNamespace) -> dict[str, Any]:
+def _get_type_hint(
+    autodoc_mock_imports: list[str], name: str, obj: Any, localns: dict[Any, MyTypeAliasForwardRef]
+) -> dict[str, Any]:
     _resolve_type_guarded_imports(autodoc_mock_imports, obj)
     try:
         result = get_type_hints(obj, None, localns)
@@ -511,7 +543,13 @@ def _get_type_hint(autodoc_mock_imports: list[str], name: str, obj: Any, localns
         else:
             result = {}
     except NameError as exc:
-        _LOGGER.warning('Cannot resolve forward reference in type annotations of "%s": %s', name, exc)
+        _LOGGER.warning(
+            'Cannot resolve forward reference in type annotations of "%s": %s',
+            name,
+            exc,
+            type="sphinx_autodoc_typehints",
+            subtype="forward_reference",
+        )
         result = obj.__annotations__
     return result
 
@@ -529,7 +567,13 @@ def backfill_type_hints(obj: Any, name: str) -> dict[str, Any]:  # noqa: C901, P
     def _one_child(module: Module) -> stmt | None:
         children = module.body  # use the body to ignore type comments
         if len(children) != 1:
-            _LOGGER.warning('Did not get exactly one node from AST for "%s", got %s', name, len(children))
+            _LOGGER.warning(
+                'Did not get exactly one node from AST for "%s", got %s',
+                name,
+                len(children),
+                type="sphinx_autodoc_typehints",
+                subtype="multiple_ast_nodes",
+            )
             return None
         return children[0]
 
@@ -554,7 +598,12 @@ def backfill_type_hints(obj: Any, name: str) -> dict[str, Any]:  # noqa: C901, P
     try:
         comment_args_str, comment_returns = type_comment.split(" -> ")
     except ValueError:
-        _LOGGER.warning('Unparseable type hint comment for "%s": Expected to contain ` -> `', name)
+        _LOGGER.warning(
+            'Unparseable type hint comment for "%s": Expected to contain ` -> `',
+            name,
+            type="sphinx_autodoc_typehints",
+            subtype="comment",
+        )
         return {}
 
     rv = {}
@@ -569,7 +618,9 @@ def backfill_type_hints(obj: Any, name: str) -> dict[str, Any]:  # noqa: C901, P
             comment_args.insert(0, None)  # self/cls may be omitted in type comments, insert blank
 
         if len(args) != len(comment_args):
-            _LOGGER.warning('Not enough type comments found on "%s"', name)
+            _LOGGER.warning(
+                'Not enough type comments found on "%s"', name, type="sphinx_autodoc_typehints", subtype="comment"
+            )
             return rv
 
     for at, arg in enumerate(args):
@@ -671,7 +722,7 @@ def process_docstring(  # noqa: PLR0913, PLR0917
     except (ValueError, TypeError):
         signature = None
 
-    localns = TypeAliasNamespace(app.config["autodoc_type_aliases"])
+    localns = {key: MyTypeAliasForwardRef(value) for key, value in app.config["autodoc_type_aliases"].items()}
     type_hints = get_all_type_hints(app.config.autodoc_mock_imports, obj, name, localns)
     app.config._annotation_globals = getattr(obj, "__globals__", {})  # noqa: SLF001
     try:
@@ -768,7 +819,10 @@ def _inject_signature(
             if annotation is None:
                 type_annotation = f":type {arg_name}: "
             else:
-                formatted_annotation = add_type_css_class(format_annotation(annotation, app.config))
+                short_literals = app.config.python_display_short_literal_types
+                formatted_annotation = add_type_css_class(
+                    format_annotation(annotation, app.config, short_literals=short_literals)
+                )
                 type_annotation = f":type {arg_name}: {formatted_annotation}"
 
             if app.config.typehints_defaults:
@@ -849,7 +903,7 @@ def get_insert_index(app: Sphinx, lines: list[str]) -> InsertIndexInfo | None:
 
     # 3. Insert after the parameters.
     # To find the parameters, parse as a docutils tree.
-    settings = OptionParser(components=(RSTParser,)).get_default_values()
+    settings = get_default_settings(RSTParser)  # type: ignore[arg-type]
     settings.env = app.env
     doc = parse("\n".join(lines), settings)
 
@@ -878,13 +932,15 @@ def get_insert_index(app: Sphinx, lines: list[str]) -> InsertIndexInfo | None:
             continue
         line_no = node_line_no(child)
         at = max(line_no - 2, 0) if line_no else len(lines)
+        if lines[at - 1]:  # skip if something on this line
+            break
         return InsertIndexInfo(insert_index=at, found_directive=True)
 
     # 5. Otherwise, insert at end
     return InsertIndexInfo(insert_index=len(lines))
 
 
-def _inject_rtype(  # noqa: PLR0913, PLR0917
+def _inject_rtype(  # noqa: C901, PLR0913, PLR0917
     type_hints: dict[str, Any],
     original_obj: Any,
     app: Sphinx,
@@ -898,6 +954,8 @@ def _inject_rtype(  # noqa: PLR0913, PLR0917
         return
     if not app.config.typehints_document_rtype:
         return
+    if not app.config.typehints_document_rtype_none and type_hints["return"] is types.NoneType:
+        return
 
     r = get_insert_index(app, lines)
     if r is None:
@@ -908,7 +966,10 @@ def _inject_rtype(  # noqa: PLR0913, PLR0917
     if not app.config.typehints_use_rtype and r.found_return and " -- " in lines[insert_index]:
         return
 
-    formatted_annotation = add_type_css_class(format_annotation(type_hints["return"], app.config))
+    short_literals = app.config.python_display_short_literal_types
+    formatted_annotation = add_type_css_class(
+        format_annotation(type_hints["return"], app.config, short_literals=short_literals)
+    )
 
     if r.found_param and insert_index < len(lines) and lines[insert_index].strip():
         insert_index -= 1
@@ -980,6 +1041,7 @@ def setup(app: Sphinx) -> dict[str, bool]:
     app.add_config_value("always_document_param_types", False, "html")  # noqa: FBT003
     app.add_config_value("typehints_fully_qualified", False, "env")  # noqa: FBT003
     app.add_config_value("typehints_document_rtype", True, "env")  # noqa: FBT003
+    app.add_config_value("typehints_document_rtype_none", True, "env")  # noqa: FBT003
     app.add_config_value("typehints_use_rtype", True, "env")  # noqa: FBT003
     app.add_config_value("typehints_defaults", None, "env")
     app.add_config_value("simplify_optional_unions", True, "env")  # noqa: FBT003
